import pathlib
from collections import OrderedDict

import anndata
import joblib
import numpy as np
import pandas as pd
import pynndescent
from scipy.cluster.hierarchy import linkage
from scipy.sparse import issparse
from scipy.stats import zscore
import scanpy as sc
import anndata as ad
from sklearn.decomposition import PCA
from sklearn.preprocessing import OneHotEncoder, normalize

from .cca import cca, lsi_cca

def scanpy_PCA_plus(
        a: ad.AnnData, n_comps: int,
        weight_by_var: bool = True, **kwargs) -> None:
    """
    Ref:
    https://github.com/satijalab/seurat/blob/1549dcb3075eaeac01c925c4b4bb73c73450fc50/R/dimensional_reduction.R#L897
    """
    sc.pp.pca(a, n_comps=n_comps, **kwargs)
    key_obsm = "X_pca"
    key_uns = "pca"
    if not weight_by_var:
        print("Normalize PCA by diving the singluar values.")
        sdev = np.sqrt(a.uns[key_uns]['variance'])
        singular_values = sdev * np.sqrt(a.shape[0] - 1)
        old_X_pca = a.obsm[key_obsm]
        new_X_pca = old_X_pca / singular_values
        a.obsm[key_obsm] = new_X_pca
        a.obsm[f"scanpy_{key_obsm}"] = old_X_pca
        a.uns[key_uns]['singular_values'] = singular_values
    else:
        print("Keep the scanpy default PCA, i.e., weighted by variance of PCs.")
    return None

def find_neighbor(cc1, cc2, k, random_state=0, n_jobs=-1):
    """
    Find all four way of neighbors for two datasets.

    Parameters
    ----------
    cc1
        cc for dataset 1
    cc2
        cc for dataset 2
    k
        number of neighbors
    random_state
        random seed
    n_jobs
        number of jobs to run in parallel

    Returns
    -------
    11, 12, 21, 22 neighbor matrix in shape (n_cell, k)
    """
    index = pynndescent.NNDescent(
        cc1,
        metric="euclidean",
        n_neighbors=k + 1,
        random_state=random_state,
        parallel_batch_queries=True,
        n_jobs=n_jobs,
    )
    G11 = index.neighbor_graph[0][:, 1: k + 1]
    G21 = index.query(cc2, k=k)[0]
    index = pynndescent.NNDescent(
        cc2,
        metric="euclidean",
        n_neighbors=k + 1,
        random_state=random_state,
        parallel_batch_queries=True,
        n_jobs=n_jobs,
    )
    G22 = index.neighbor_graph[0][:, 1: k + 1]
    G12 = index.query(cc1, k=k)[0]
    return G11, G12, G21, G22


def find_mnn(G12, G21, kanchor):
    """Calculate mutual nearest neighbor for two datasets."""
    anchor = [
        [i, G12[i, j]]
        for i in range(G12.shape[0])
        for j in range(kanchor)
        if (i in G21[G12[i, j], :kanchor])
    ]
    return np.array(anchor)


def min_max(tmp, q_left=1, q_right=90):
    """Normalize to q_left, q_right quantile to 0, 1, and cap extreme values."""
    tmin, tmax = np.percentile(tmp, [q_left, q_right])
    tmp = (tmp - tmin) / (tmax - tmin)
    tmp[tmp > 1] = 1
    tmp[tmp < 0] = 0
    return tmp


def filter_anchor(
    anchor,
    adata_ref=None,
    adata_qry=None,
    scale_ref=False,
    scale_qry=False,
    high_dim_feature=None,
    k_filter=200,
    random_state=0,
    n_jobs=-1,
):
    """
    Check if an anchor is still an anchor when only using the high_dim_features to construct KNN graph.

    If not, remove the anchor.
    """
    if issparse(adata_ref.X):
        ref_data = adata_ref.X[:, high_dim_feature].toarray()
    else:
        ref_data = adata_ref.X[:, high_dim_feature].copy()
    if scale_ref:
        ref_data = zscore(ref_data, axis=0)
    ref_data = normalize(ref_data, axis=1)

    if issparse(adata_qry.X):
        qry_data = adata_qry.X[:, high_dim_feature].toarray()
    else:
        qry_data = adata_qry.X[:, high_dim_feature].copy()
    if scale_qry:
        qry_data = zscore(qry_data, axis=0)
    qry_data = normalize(qry_data, axis=1)

    index = pynndescent.NNDescent(
        ref_data,
        metric="euclidean",
        n_neighbors=k_filter,
        random_state=random_state,
        parallel_batch_queries=True,
        n_jobs=n_jobs,
    )
    G = index.query(qry_data, k=k_filter)[0]
    input_anchors = anchor.shape[0]
    anchor = np.array([xx for xx in anchor if (xx[0] in G[xx[1]])])
    print(
        f"Anchor selected with high CC feature graph: {anchor.shape[0]} / {input_anchors}"
    )
    return anchor


def score_anchor(
    anchor, G11, G12, G21, G22, k_score=30, Gp1=None, Gp2=None, k_local=50
):
    """
    Score the anchor by the number of shared neighbors.

    Parameters
    ----------
    anchor
        anchor in shape (n_anchor, 2)
    G11
        neighbor graph of dataset 1
    G12
        neighbor graph of dataset 1 to 2
    G21
        neighbor graph of dataset 2 to 1
    G22
        neighbor graph of dataset 2
    k_score
        number of neighbors to score the anchor
    Gp1
        Intra-dataset1 kNN graph
    Gp2
        Intra-dataset2 kNN graph
    k_local
        number of neighbors to calculate the local score

    Returns
    -------
    anchor with score in shape (n_anchor, 3): pd.DataFrame
    """
    tmp = [
        len(set(G11[x, :k_score]).intersection(G21[y, :k_score]))
        + len(set(G12[x, :k_score]).intersection(G22[y, :k_score]))
        for x, y in anchor
    ]
    anchor_df = pd.DataFrame(anchor, columns=["x1", "x2"])
    anchor_df["score"] = min_max(tmp)

    if k_local:
        # if k_local is not None, then use local KNN to adjust the score
        share_nn = np.array(
            [len(set(Gp1[i]).intersection(G11[i, :k_local])) for i in range(len(Gp1))]
        )
        tmp = [share_nn[xx] for xx in anchor_df["x1"].values]
        anchor_df["score_local1"] = min_max(tmp)

        share_nn = np.array(
            [len(set(Gp2[i]).intersection(G22[i, :k_local])) for i in range(len(Gp2))]
        )
        tmp = [share_nn[xx] for xx in anchor_df["x2"].values]
        anchor_df["score_local2"] = min_max(tmp)

        anchor_df["score"] = (
            anchor_df["score"] * anchor_df["score_local1"] * anchor_df["score_local2"]
        )
    return anchor_df


def find_order(dist, ncell):
    """Use dendrogram to find the order of dataset pairs."""
    D = linkage(1 / dist, method="average")
    node_dict = {i: [i] for i in range(len(ncell))}
    alignment = []
    for xx in D[:, :2].astype(int):
        if ncell[xx[0]] < ncell[xx[1]]:
            xx = xx[::-1]
        alignment.append([node_dict[xx[0]], node_dict[xx[1]]])
        node_dict[len(ncell)] = node_dict[xx[0]] + node_dict[xx[1]]
        ncell.append(ncell[xx[0]] + ncell[xx[1]])
    return alignment


class SeuratIntegration:
    """Main class for Seurat integration."""

    def __init__(self, n_jobs=-1, random_state=0):
        self.n_jobs = n_jobs

        # intra-dataset KNN graph
        self.k_local = None
        self.key_local = None
        self.local_knn = []

        self.adata_dict = OrderedDict()
        self.n_dataset = 0
        self.n_cells = []
        self.alignments = None
        self.all_pairs = np.array([])
        self._get_all_pairs()

        self.anchor = {}
        self.mutual_knn = {}
        self.raw_anchor = {}
        self.label_transfer_results = {}

        self.random_state = random_state

    def _calculate_local_knn(self):
        """
        Calculate local kNN graph for each dataset.

        If klocal is provided, we calculate the local knn graph to
        evaluate whether the anchor preserves local structure within the dataset.
        One can use a different obsm with key_local to compute knn for each dataset.
        """
        if self.k_local is not None:
            print("Find neighbors within datasets")
            for adata in self.adata_dict.values():
                index = pynndescent.NNDescent(
                    adata.obsm[self.key_local],
                    metric="euclidean",
                    n_neighbors=self.k_local + 1,
                    random_state=self.random_state,
                    parallel_batch_queries=True,
                    n_jobs=self.n_jobs,
                )
                self.local_knn.append(index.neighbor_graph[0][:, 1:])
        else:
            self.local_knn = [None for _ in self.adata_dict.values()]

    def _get_all_pairs(self):
        if self.alignments is not None:
            all_pairs = []
            for pair in self.alignments:
                for xx in pair[0]:
                    for yy in pair[1]:
                        if xx < yy:
                            all_pairs.append(f"{xx}-{yy}")
                        else:
                            all_pairs.append(f"{yy}-{xx}")
            self.all_pairs = np.unique(all_pairs)
        else:
            self.all_pairs = np.array([])

    def _prepare_matrix(self, i, j, key_anchor):
        adata_dict = self.adata_dict
        adata1 = adata_dict[i]
        adata2 = adata_dict[j]

        if key_anchor == "X":
            # in case the adata var is not in the same order
            # select and order the var to make sure it is matched
            if (adata1.shape[1] != adata2.shape[1]) or (
                (adata1.var.index == adata2.var.index).sum() < adata1.shape[1]
            ):
                sel_b = adata1.var.index & adata2.var.index
                U = adata1[:, sel_b].X.copy()
                V = adata2[:, sel_b].X.copy()
            else:
                U = adata1.X.copy()
                V = adata2.X.copy()
        else:
            U = adata1.obsm[key_anchor]
            V = adata2.obsm[key_anchor]

        return U, V

    def _calculate_mutual_knn_and_raw_anchors(self, i, j, U, V, k, k_anchor):
        """
        Calculate the mutual knn graph and raw anchors.

        The results are saved to self.mutual_knn and self.raw_anchor.
        """
        G11, G12, G21, G22 = find_neighbor(U, V, k=k, n_jobs=self.n_jobs)
        raw_anchors = find_mnn(G12, G21, k_anchor)
        self.mutual_knn[(i, j)] = (G11, G12, G21, G22)
        self.raw_anchor[(i, j)] = raw_anchors
        return G11, G12, G21, G22, raw_anchors

    def _pairwise_find_anchor(
        self,
        i,
        i_sel,
        j,
        j_sel,
        dim_red,
        key_anchor,
        svd_algorithm,
        scale1,
        scale2,
        k_anchor,
        k_local,
        k_score,
        ncc,
        max_cc_cell,
        k_filter,
        n_features,
        chunk_size,
        random_state,
        signorm,
    ):
        """Pairwise anchor between two datasets."""
        adata1 = self.adata_dict[i]
        adata2 = self.adata_dict[j]

        min_sample = min(adata1.shape[0], adata2.shape[0])

        if i_sel is not None:
            adata1 = adata1[i_sel, :]
        if j_sel is not None:
            adata2 = adata2[j_sel, :]
        if dim_red in ("cca", "pca", "lsi", "lsi-cca"):
            print(f"1. Prepare input matrix using {key_anchor}.")
            U, V = self._prepare_matrix(i, j, key_anchor=key_anchor)
            if dim_red in ("cca", "pca"):
                print("2. Run CCA")
                U, V, high_dim_feature = cca(
                    data1=U,
                    data2=V,
                    scale1=scale1,
                    scale2=scale2,
                    n_components=ncc,
                    max_cc_cell=max_cc_cell,
                    k_filter=k_filter,
                    n_features=n_features,
                    chunk_size=chunk_size,
                    svd_algorithm=svd_algorithm,
                    random_state=random_state,
                )
            else:
                print("2. Run LSI-CCA")
                U, V = lsi_cca(
                    data1=U,
                    data2=V,
                    scale_factor=100000,
                    n_components=ncc,
                    max_cc_cell=max_cc_cell,
                    chunk_size=chunk_size,
                    svd_algorithm=svd_algorithm,
                    min_cov_filter=5,
                    random_state=random_state,
                )
                high_dim_feature = None
            print("3. Normalize CCV per sample/row")
            U = normalize(U, axis=1)
            V = normalize(V, axis=1)
            print("4. find MNN of U and V to get anchors.")
            _k = max(_temp for _temp
                    in [k_anchor, k_local, k_score] if _temp is not None)
            _k = min(min_sample - 2, _k)
            print(f"Find Anchors using k={_k}")
            G11, G12, G21, G22, raw_anchors = (
                self._calculate_mutual_knn_and_raw_anchors(
                    i=i, j=j, U=U, V=V, k=_k, k_anchor=k_anchor
                )
            )

            print("5. filter anchors by high dimensional neighbors.")
            # compute ccv feature loading
            if k_filter is not None and high_dim_feature is not None:
                if self.n_cells[i] >= self.n_cells[j]:
                    raw_anchors = filter_anchor(
                        anchor=raw_anchors,
                        adata_ref=adata1,
                        adata_qry=adata2,
                        scale_ref=scale1,
                        scale_qry=scale2,
                        high_dim_feature=high_dim_feature,
                        k_filter=k_filter,
                        random_state=self.random_state,
                        n_jobs=self.n_jobs,
                    )
                else:
                    raw_anchors = filter_anchor(
                        anchor=raw_anchors[:, ::-1],
                        adata_ref=adata2,
                        adata_qry=adata1,
                        scale_ref=scale2,
                        scale_qry=scale1,
                        high_dim_feature=high_dim_feature,
                        k_filter=k_filter,
                        random_state=self.random_state,
                        n_jobs=self.n_jobs,
                    )[:, ::-1]
        elif dim_red in ("rpca", "rlsi"):
            print(f"Perform {dim_red}.")
            from .cca import LSI, SVD, downsample

            adata1, adata2 = adata1.X, adata2.X
            k = max(i for i in [k_anchor, k_local, k_score, 50] if i is not None)
            if dim_red == "rpca":
                print("Run rPCA")
                model = SVD(n_components=ncc, random_state=random_state)
            elif dim_red == "rlsi":
                print("Run rLSI")
                model = LSI(n_components=ncc, random_state=random_state)
            else:
                raise ValueError(
                    f"Dimension reduction method {dim_red} is not supported."
                )
            tf1, tf2, scaler1, scaler2 = downsample(
                adata1,
                adata2,
                todense=True if dim_red == "rpca" else False,
                scale1=scale1,
                scale2=scale2,
                max_cc_cell=max_cc_cell,
            )

            # project adata2 to adata1
            model.fit(tf1)
            U = model.transform(adata1, chunk_size=chunk_size, scaler=scaler1)
            V = model.transform(adata2, chunk_size=chunk_size, scaler=scaler2)
            if (dim_red == "pca") and signorm:
                U = U / model.model.singular_values_
                V = V / model.model.singular_values_
            index = pynndescent.NNDescent(
                U,
                metric="euclidean",
                n_neighbors=k + 1,
                random_state=random_state,
                n_jobs=-1,
            )
            G11 = index.neighbor_graph[0][:, 1: k + 1]
            G21 = index.query(V, k=k)[0]

            # project adata1 to adata2
            model.fit(tf2)
            U = model.transform(adata1, chunk_size=chunk_size, scaler=scaler1)
            V = model.transform(adata2, chunk_size=chunk_size, scaler=scaler2)
            if (dim_red == "pca") and signorm:
                U = U / model.model.singular_values_
                V = V / model.model.singular_values_
            index = pynndescent.NNDescent(
                V,
                metric="euclidean",
                n_neighbors=k + 1,
                random_state=random_state,
                n_jobs=-1,
            )
            G22 = index.neighbor_graph[0][:, 1: k + 1]
            G12 = index.query(U, k=k)[0]

            raw_anchors = find_mnn(G12, G21, k_anchor)
        else:
            raise ValueError(f"{dim_red} is not supported.")

        print("6. Score anchors with snn and local structure preservation.")
        anchor_df = score_anchor(
            anchor=raw_anchors,
            G11=G11,
            G12=G12,
            G21=G21,
            G22=G22,
            k_score=k_score,
            k_local=k_local,
            Gp1=self.local_knn[i],
            Gp2=self.local_knn[j])
        return anchor_df

    def find_anchor(
        self,
        adata_list,
        adata_names=None,
        k_local=None,
        key_local="X_pca",
        key_anchor="X",
        dim_red="pca",
        svd_algorithm="randomized",
        scale1=True,
        scale2=True,
        scale_list=None,
        k_filter=None,
        n_features=200,
        n_components=None,
        max_cc_cells=50000,
        chunk_size=50000,
        k_anchor=5,
        k_score=30,
        alignments=None,
        random_state=0,
        signorm=True,
        key_match=None,
    ):
        """Find anchors for each dataset pair.

        Parameters
        ----------
        key_local
            str, used for KNN construction in each dataset
        key_anchor
            str, used for anchor construction
        dim_red
            str from pca, cca, lsi, lsi-cca, rpca, rlsi
            Based on current implementation,
            - choosing pca or cca, will perform CCA
            - choosing lsi or lsi-cca, will perform LSI-CCA
            - rpca is only used when key_match is None at least
        n_components
            int, dim used for CCA and other methods, default None will set it as 50.
        adata_names
            list of int.
        """
        valid_dim_red_name = ["pca", "cca", "lsi", "lsi-cca", "rpca", "rlsi"]
        if dim_red not in valid_dim_red_name:
            raise ValueError(f"Dimension reduction method {dim_red} is not supported.")

        if adata_names is None:
            adata_names = list(range(len(adata_list)))
        try:
            assert len(adata_names) == len(adata_list)
        except AssertionError:
            print("length of adata_names does not match length of adata_list.")

        self.adata_dict = {k: v for k, v in zip(adata_names, adata_list)}
        self.n_dataset = len(adata_list)
        self.n_cells = [adata.shape[0] for adata in adata_list]

        # intra-dataset KNN for scoring the anchors
        self.k_local = k_local
        self.key_local = key_local
        self._calculate_local_knn()

        # alignments and all_pairs
        self.alignments = alignments
        self._get_all_pairs()

        print("Find anchors across datasets.")
        for i in range(self.n_dataset - 1):
            for j in range(i + 1, self.n_dataset):
                if scale_list is not None:
                    scale1 = scale_list[i]
                    scale2 = scale_list[j]
                    print("Get scale1 and scale2 from scale_list")
                    print(f"dataset {i} scale: {scale1}")
                    print(f"dataset {j} scale: {scale2}")

                if key_match is None:
                    anchor_df = self._pairwise_find_anchor(
                        i=i,
                        i_sel=None,
                        j=j,
                        j_sel=None,
                        dim_red=dim_red,
                        key_anchor=key_anchor,
                        svd_algorithm=svd_algorithm,
                        scale1=scale1,
                        scale2=scale2,
                        k_anchor=k_anchor,
                        k_local=k_local,
                        k_score=k_score,
                        ncc=n_components,
                        max_cc_cell=max_cc_cells,
                        k_filter=k_filter,
                        n_features=n_features,
                        chunk_size=chunk_size,
                        random_state=random_state,
                        signorm=signorm,
                    )
                else:
                    tissue = [xx.obs[key_match].unique() for xx in adata_list]
                    sharet = list(set(tissue[i]).intersection(tissue[j]))
                    if len(sharet) > 0:
                        anchor_df_list = []
                        for t in sharet:
                            print(t)
                            adata1 = adata_list[i].copy()
                            adata2 = adata_list[j].copy()

                            idx1 = np.where(adata1.obs[key_match] == t)[0]
                            idx2 = np.where(adata2.obs[key_match] == t)[0]
                            tmp = self._pairwiser_find_anchor(
                                i=i,
                                i_sel=idx1,
                                j=j,
                                j_sel=idx2,
                                dim_red=dim_red,
                                key_anchor=key_anchor,
                                svd_algorithm=svd_algorithm,
                                scale1=scale1,
                                scale2=scale2,
                                k_anchor=k_anchor,
                                k_local=k_local,
                                k_score=k_score,
                                ncc=n_components,
                                max_cc_cell=max_cc_cells,
                                k_filter=k_filter,
                                n_features=n_features,
                                chunk_size=chunk_size,
                                random_state=random_state,
                                signorm=signorm,
                            )
                            tmp["x1"] = idx1[tmp["x1"].values]
                            tmp["x2"] = idx2[tmp["x2"].values]
                            anchor_df_list.append(tmp)
                        anchor_df = pd.concat(anchor_df_list, axis=0)
                    else:
                        anchor_df = self._pairwise_find_anchor(
                            i=i,
                            i_sel=None,
                            j=j,
                            j_sel=None,
                            dim_red="rpca",
                            key_anchor=key_anchor,
                            svd_algorithm=svd_algorithm,
                            scale1=scale1,
                            scale2=scale2,
                            k_anchor=k_anchor,
                            k_local=k_local,
                            k_score=k_score,
                            ncc=n_components,
                            max_cc_cell=max_cc_cells,
                            k_filter=k_filter,
                            n_features=n_features,
                            chunk_size=chunk_size,
                            random_state=random_state,
                            signorm=signorm,
                        )

                # save anchors
                self.anchor[(i, j)] = anchor_df.copy()
                print(
                    f"Identified {len(self.anchor[i, j])} anchors between datasets {i} and {j}."
                )
        return

    def find_nearest_anchor(
        self,
        data,
        data_qry,
        ref,
        qry,
        key_correct="X_pca",
        npc=30,
        k_weight=100,
        sd=1,
        random_state=0,
    ):
        """Find the nearest anchors for each cell in data."""
        print("Initialize")
        cum_ref, cum_qry = [0], [0]
        for xx in ref:
            cum_ref.append(cum_ref[-1] + data[xx].shape[0])
        for xx in qry:
            cum_qry.append(cum_qry[-1] + data[xx].shape[0])

        anchor = []
        for i, xx in enumerate(ref):
            for j, yy in enumerate(qry):
                if xx < yy:
                    tmp = self.anchor[(xx, yy)].copy()
                else:
                    tmp = self.anchor[(yy, xx)].copy()
                    tmp[["x1", "x2"]] = tmp[["x2", "x1"]]
                tmp["x1"] += cum_ref[i]
                tmp["x2"] += cum_qry[j]
                anchor.append(tmp)
        anchor = pd.concat(anchor)
        score = anchor["score"].values
        anchor = anchor[["x1", "x2"]].values

        if key_correct == "X":
            model = PCA(
                n_components=npc, svd_solver="arpack", random_state=random_state
            )
            reduce_qry = model.fit_transform(data_qry)
        else:
            reduce_qry = data_qry[:, :npc]

        print("Find nearest anchors", end=". ")
        index = pynndescent.NNDescent(
            reduce_qry[anchor[:, 1]],
            metric="euclidean",
            n_neighbors=k_weight,
            random_state=random_state,
            parallel_batch_queries=True,
            n_jobs=self.n_jobs,
        )
        k_weight = min(k_weight, anchor.shape[0] - 5)
        k_weight = max(5, k_weight)
        print("k_weight: ", k_weight, end="\n")
        G, D = index.query(reduce_qry, k=k_weight)

        print("Normalize graph")
        cell_filter = D[:, -1] == 0
        D = (1 - D / D[:, -1][:, None]) * score[G]
        D[cell_filter] = score[G[cell_filter]]
        D = 1 - np.exp(-D * (sd**2) / 4)
        D = D / (np.sum(D, axis=1) + 1e-6)[:, None]
        return anchor, G, D, cum_qry

    def transform(
        self,
        data,
        ref,
        qry,
        key_correct,
        npc=30,
        k_weight=100,
        sd=1,
        chunk_size=50000,
        random_state=0,
        row_normalize=True,
    ):
        """Transform query data to reference space."""
        data_ref = np.concatenate(data[ref])
        data_qry = np.concatenate(data[qry])

        anchor, G, D, cum_qry = self.find_nearest_anchor(
            data=data,
            data_qry=data_qry,
            key_correct=key_correct,
            ref=ref,
            qry=qry,
            npc=npc,
            k_weight=k_weight,
            sd=sd,
            random_state=random_state,
        )

        print("Transform data")
        bias = data_ref[anchor[:, 0]] - data_qry[anchor[:, 1]]
        data_prj = np.zeros(data_qry.shape)

        for chunk_start in np.arange(0, data_prj.shape[0], chunk_size):
            data_prj[chunk_start: (chunk_start + chunk_size)] = data_qry[
                chunk_start: (chunk_start + chunk_size)
            ] + (
                D[chunk_start: (chunk_start + chunk_size), :, None]
                * bias[G[chunk_start: (chunk_start + chunk_size)]]
            ).sum(
                axis=1
            )
        for i, xx in enumerate(qry):
            _data = data_prj[cum_qry[i]: cum_qry[i + 1]]
            if row_normalize:
                _data = normalize(_data, axis=1)
            data[xx] = _data
        return data

    def integrate(
        self,
        key_correct,
        row_normalize=True,
        n_components=30,
        k_weight=100,
        sd=1,
        alignments=None,
    ):
        """\
        Map query data to reference space.

        Transform query matrices datasets by transform data matrices
        from query to reference data using the MNN information.
        Reference data will not be changed or perform L2-normalization
        per row if row_normalize = True.
        
        NOTE: From the implementation, it will go through each element in
        the alignments, but only return the last one. So we should
        consider the ref and quey pair for this function.
        
        Parameters
        ----------
        key_correct
            str, X or field of obsm, such as X_pca
        row_normalize
            boolean, if perform L2-normalized for key_correct
            default is True

        Returns
        -------
        List of numpy arrays, ordered by alignments (ref, qeury).
        """
        if alignments is not None:
            self.alignments = alignments

        # find order of pairwise dataset merging with hierarchical clustering
        if self.alignments is None:
            dist = []
            for i in range(self.n_dataset - 1):
                for j in range(i + 1, self.n_dataset):
                    dist.append(
                        len(self.anchor[(i, j)])
                        / min([self.n_cells[i], self.n_cells[j]])
                    )
            self.alignments = find_order(np.array(dist), self.n_cells)
            print(f"Alignments: {self.alignments}")

        print("Merge datasets")
        adata_list = list(self.adata_dict.values())

        # initialize corrected with original data
        if key_correct == "X":
            # correct the original feature matrix
            corrected = [adata_list[i].X.copy() for i in range(self.n_dataset)]
        else:
            # correct dimensionality reduced matrix only
            if row_normalize:
                corrected = [
                    normalize(adata_list[i].obsm[key_correct], axis=1)
                    for i in range(self.n_dataset)]
            else:
                corrected = [
                    adata_list[i].obsm[key_correct] for i in range(self.n_dataset)]

        for xx in self.alignments:
            print(xx)
            corrected = self.transform(
                data=np.array(corrected, dtype="object"),
                ref=xx[0],
                qry=xx[1],
                npc=n_components,
                k_weight=k_weight,
                sd=sd,
                random_state=self.random_state,
                row_normalize=row_normalize,
                key_correct=key_correct,
            )
        return corrected

    def label_transfer(
        self,
        ref,
        qry,
        categorical_key=None,
        continuous_key=None,
        key_dist="X_pca",
        k_weight=100,
        npc=30,
        sd=1,
        chunk_size=50000,
        random_state=0,
    ):
        """Transfer labels from query to reference space."""
        adata_list = list(self.adata_dict.values())

        data_qry = np.concatenate(
            [normalize(adata_list[i].obsm[key_dist], axis=1) for i in qry]
        )
        data_qry_index = np.concatenate([adata_list[i].obs_names for i in qry])

        anchor, G, D, cum_qry = self.find_nearest_anchor(
            data=adata_list,
            data_qry=data_qry,
            ref=ref,
            qry=qry,
            npc=npc,
            k_weight=k_weight,
            key_correct=key_dist,
            sd=sd,
            random_state=random_state,
        )
        print("Label transfer")
        label_ref = []
        columns = []
        cat_counts = []

        if categorical_key is None:
            categorical_key = []
        if continuous_key is None:
            continuous_key = []
        if len(categorical_key) == 0 and len(continuous_key) == 0:
            raise ValueError("No categorical or continuous key specified.")

        if len(categorical_key) > 0:
            tmp = pd.concat([adata_list[i].obs[categorical_key] for i in ref], axis=0)
            enc = OneHotEncoder()
            label_ref.append(
                enc.fit_transform(tmp[categorical_key].values.astype(np.str_)).toarray()
            )
            # add categorical key to make sure col is unique
            columns += enc.categories_
            # enc.categories_ are a list of arrays, each array are categories in that categorical_key
            cat_counts += [cats.size for cats in enc.categories_]

        if len(continuous_key) > 0:
            tmp = pd.concat([adata_list[i].obs[continuous_key] for i in ref], axis=0)
            label_ref.append(tmp[continuous_key].values)
            columns += [[xx] for xx in continuous_key]
            cat_counts += [1 for _ in continuous_key]

        label_ref = np.concatenate(label_ref, axis=1)
        label_qry = np.zeros((data_qry.shape[0], label_ref.shape[1]))

        bias = label_ref[anchor[:, 0]]
        for chunk_start in np.arange(0, label_qry.shape[0], chunk_size):
            label_qry[chunk_start: (chunk_start + chunk_size)] = (
                D[chunk_start: (chunk_start + chunk_size), :, None]
                * bias[G[chunk_start: (chunk_start + chunk_size)]]
            ).sum(axis=1)

        # these column names might be duplicated
        all_column_names = np.concatenate(columns)
        all_column_variables = np.repeat(categorical_key + continuous_key, cat_counts)
        label_qry = pd.DataFrame(
            label_qry, index=data_qry_index, columns=all_column_names
        )
        result = {}
        for key in categorical_key + continuous_key:
            result[key] = label_qry.iloc[:, all_column_variables == key]
        return result

    def save(
        self,
        output_path,
        save_local_knn=False,
        save_raw_anchor=False,
        save_mutual_knn=False,
        save_adata=False,
    ):
        """Save the model and results to disk."""
        # save each adata in a separate dir
        output_path = pathlib.Path(output_path)
        output_path.mkdir(exist_ok=True, parents=True)

        if save_adata:
            # save adata and clear the self.adata_dict
            adata_dir = output_path / "adata"
            adata_dir.mkdir(exist_ok=True)
            with open(f"{adata_dir}/order.txt", "w") as f:
                for k, v in self.adata_dict.items():
                    for col, val in v.obs.items():
                        if val.dtype == "O":
                            v.obs[col] = val.fillna("nan").astype(str)
                        elif val.dtype == "category":
                            v.obs[col] = val.fillna("nan").astype(str)
                        else:
                            pass
                    v.write_h5ad(f"{adata_dir}/{k}.h5ad")
                    f.write(f"{k}\n")

        # clear the adata in integrator
        self.adata_dict = {}

        if not save_local_knn:
            self.local_knn = []
        if not save_raw_anchor:
            self.raw_anchor = {}
        if not save_mutual_knn:
            self.mutual_knn = {}

        joblib.dump(self, f"{output_path}/model.lib")
        return

    @classmethod
    def load(cls, input_path):
        """Load integrator from file."""
        adata_dir = f"{input_path}/adata"
        model_path = f"{input_path}/model.lib"

        obj = joblib.load(model_path)

        orders = pd.read_csv(f"{adata_dir}/order.txt", header=None, index_col=0).index
        adata_dict = OrderedDict()
        for k in orders:
            adata_path = f"{adata_dir}/{k}.h5ad"
            if pathlib.Path(adata_path).exists():
                adata_dict[k] = anndata.read_h5ad(f"{adata_dir}/{k}.h5ad")
        obj.adata_dict = adata_dict
        return obj

    @classmethod
    def save_transfer_results_to_adata(
        cls, adata, transfer_results, new_label_suffix="_transfer"
    ):
        """Save transfer results to adata."""
        for key, df in transfer_results.items():
            adata.obs[key + new_label_suffix] = adata.obs[key].copy()
            adata.obs.loc[df.index, key + new_label_suffix] = df.idxmax(axis=1).values
        return



from ._version import __version__
from . import metrics

import sys

import logging
logging.basicConfig(
    stream=sys.stderr,
    format="%(asctime)s - %(levelname)s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    level=logging.INFO, 
)

del sys


from __future__ import annotations

from typing import Literal
import scipy as sp
import numpy as np
import gc
import logging
import math

from snapatac2._utils import is_anndata 
import snapatac2._snapatac2 as internal

__all__ = ['umap', 'spectral', 'multi_spectral']

def umap(
    adata: internal.AnnData | internal.AnnDataSet | np.ndarray,
    n_comps: int = 2,
    use_dims: int | list[int] | None = None,
    use_rep: str = "X_spectral",
    key_added: str = 'umap',
    random_state: int | None = 0,
    inplace: bool = True,
    **kwargs
) -> np.ndarray | None:
    """
    Parameters
    ----------
    adata
        The annotated data matrix.
    n_comps
        The number of dimensions of the embedding.
    use_dims
        Use these dimensions in `use_rep`.
    use_rep
        Use the indicated representation in `.obsm`.
    key_added
        `adata.obs` key under which to add the cluster labels.
    random_state
        Random seed.
    inplace
        Whether to store the result in the anndata object.
    **kwargs
        Other parameters defined in umap.UMAP

    Returns
    -------
    np.ndarray | None
        if `inplace=True` it stores UMAP embedding in
        `adata.obsm["X_`key_added`"]`.
        Otherwise, it returns the result as a numpy array.
    """
    from umap import UMAP

    if is_anndata(adata):
        data = adata.obsm[use_rep]
    else:
        data = adata
        inplace = False

    if use_dims is not None:
        data = data[:, :use_dims] if isinstance(use_dims, int) else data[:, use_dims]

    umap = UMAP(random_state=random_state,
                n_components=n_comps,
                **kwargs).fit_transform(data)
    if inplace:
        adata.obsm["X_" + key_added] = umap
    else:
        return umap

def idf(data, features=None):
    n, m = data.shape
    count = np.zeros(m)
    for batch, _, _ in data.chunked_X(2000):
        batch.data = np.ones(batch.indices.shape, dtype=np.float64)
        count += np.ravel(batch.sum(axis = 0))
    if features is not None:
        count = count[features]
    return np.log(n / (1 + count))

def spectral(
    adata: internal.AnnData | internal.AnnDataSet,
    n_comps: int = 30,
    features: str | np.ndarray | None = "selected",
    random_state: int = 0,
    sample_size: int | float | None = None,
    sample_method: Literal["random", "degree"] = "random",
    chunk_size: int = 5000,
    distance_metric: Literal["jaccard", "cosine"] = "cosine",
    weighted_by_sd: bool = True,
    feature_weights: list[float] | None = None,
    inplace: bool = True,
    num_threads: int = 32,
) -> tuple[np.ndarray, np.ndarray] | None:
    """
    Perform dimension reduction using Laplacian Eigenmaps.

    Convert the cell-by-feature count matrix into lower dimensional representations
    using the spectrum of the normalized graph Laplacian defined by pairwise similarity
    between cells.
    This function utilizes the matrix-free spectral embedding algorithm to compute
    the embedding when `distance_metric` is "cosine", which scales linearly with the
    number of cells. For other types of similarity metrics, the time and space complexity
    scale quadratically with the number of cells.
    
    Note
    ----
    - Determining the appropriate number of components is crucial when performing
      downstream analyses to ensure optimal clustering outcomes. Utilizing components
      that are either uninformative or irrelevant can compromise the quality of the results.
      By default, this function adopts a strategy where all eigenvectors are weighted
      according to the square root of their corresponding eigenvalues, rather than
      implementing a strict cutoff threshold. This method generally provides satisfactory
      results, circumventing the necessity for manual specification of component numbers.
      However, it's important to note that there might be exceptional cases with
      certain datasets where deviating from this default setting could yield better
      outcomes. In such scenarios, you can disable the automatic weighting by
      setting `weighted_by_sd=False`. Subsequently, you will need to manually determine
      and select the number of components to use for your specific analysis.
    - This funciton may not always return the exact number of eigenvectors requested.
      This function computes lower-dimensional embeddings by performing the
      eigen-decomposition of the normalized graph Laplacian matrix, where all
      eigenvalues should be non-negative. However, the method used to calculate
      eigenvectors, specifically `scipy.sparse.linalg.eigsh`, may not perform
      optimally for small eigenvalues. This occasionally leads to the function
      outputting negative eigenvalues at the lower spectrum. To address this issue,
      a post-processing step is introduced to eliminate these erroneous eigenvalues
      when `weighted_by_sd=True` (which is the default setting). This step
      typically has minimal impact, as the affected eigenvalues are generally very small.

    Parameters
    ----------
    adata
        AnnData or AnnDataSet object.
    n_comps
        Number of dimensions to keep. The result is insensitive to this parameter when
        `weighted_by_sd` is set to True, as long as it is large enough, e.g. 30.
    features
        Boolean index mask. True means that the feature is kept.
        False means the feature is removed. If `features=None`, all features are used.
    random_state
        Seed of the random state generator
    sample_size
        Approximate the embedding using the Nystrom algorithm by selecting
        a subset of cells. It could be either an integer
        indicating the number of cells to sample or a real value from 0 to 1
        indicating the fraction of cells to sample. When `sample_size` is None,
        the full matrix is used. Using this only when the number of cells is too large, e.g. > 10,000,000, or
        the `distance_metric` is "jaccard".
    chunk_size
        Chunk size used in the Nystrom method. The effective chunk size is
        `chunk_size` x `num_threads`. This parameter should not be too small, e.g., <1000.
    distance_metric
        distance metric: "jaccard", "cosine".
        When "cosine" is used, the matrix-free spectral embedding algorithm is used.
    weighted_by_sd
        Whether to weight the result eigenvectors by the square root of eigenvalues.
        This parameter is turned on by default. When it is turned on, mannully selecting
        the number of components is usually not necessary.
    feature_weights
        Feature weights used in the distance metric. If None, the inverse document
        frequency (IDF) is used.
    inplace
        Whether to store the result in the anndata object.
    num_threads
        Number of threads to use in the Nystrom method.

    Returns
    -------
    tuple[np.ndarray, np.ndarray] | None
        if `inplace=True` it stores Spectral embedding of data in
        `adata.obsm["X_spectral"]` and `adata.uns["spectral_eigenvalue"]`.
        Otherwise, it returns the result as numpy arrays.

    See Also
    --------
    multi_spectral
    """
    np.random.seed(random_state)

    if isinstance(features, str):
        if features in adata.var:
            features = adata.var[features].to_numpy()
        else:
            raise NameError("Please call `select_features` first or explicitly set `features = None`")

    n_comps = min(adata.n_vars - 1, adata.n_obs - 1, n_comps)

    n_sample, _ = adata.shape
    if sample_size is None:
        sample_size = n_sample
    elif isinstance(sample_size, int):
        if sample_size <= 1:
            raise ValueError("when sample_size is an integer, it should be > 1")
        if sample_size > n_sample:
            sample_size = n_sample
    else:
        if sample_size <= 0.0 or sample_size > 1.0:
            raise ValueError("when sample_size is a float, it should be > 0 and <= 1")
        else:
            sample_size = int(sample_size * n_sample)

    if sample_size >= n_sample:
        if distance_metric == "cosine":
            evals, evecs = internal.spectral_embedding(adata, features, n_comps, random_state, feature_weights)
        else:
            if feature_weights is None:
                feature_weights = idf(adata, features)
            model = Spectral(n_comps, distance_metric, feature_weights)
            X = adata.X[...] if features is None else adata.X[:, features]
            model.fit(X)
            evals, evecs = model.transform()
    else:
        logging.info("Perform spectral embedding using the Nystrom algorithm...")
        if distance_metric == "cosine":
            if sample_method == "random":
                weighted_by_degree = False
            else:
                weighted_by_degree = True
            v, u = internal.spectral_embedding_nystrom(adata, features, n_comps, sample_size, weighted_by_degree, chunk_size, None, num_threads)
            evals, evecs = orthogonalize(v, u)
        else:
            if feature_weights is None:
                feature_weights = idf(adata, features)
            model = Spectral(n_comps, distance_metric, feature_weights)
            if adata.isbacked:
                S = adata.X.chunk(sample_size, replace=False)
            else:
                S = sp.sparse.csr_matrix(adata.chunk_X(sample_size, replace=False))
            if features is not None: S = S[:, features]

            model.fit(S)

            from tqdm import tqdm
            for batch, _, _ in tqdm(adata.chunked_X(chunk_size), total=math.ceil(adata.n_obs/chunk_size)):
                if distance_metric == "jaccard":
                    batch.data = np.ones(batch.indices.shape, dtype=np.float64)
                if features is not None: batch = batch[:, features]
                model.extend(batch)
            evals, evecs = model.transform()

    if weighted_by_sd:
        idx = [i for i in range(evals.shape[0]) if evals[i] > 0]
        evals = evals[idx]
        evecs = evecs[:, idx] * np.sqrt(evals)

    if inplace:
        adata.uns['spectral_eigenvalue'] = evals
        adata.obsm['X_spectral'] = evecs
    else:
        return (evals, evecs)


class Spectral:
    def __init__(
        self,
        out_dim: int = 30,
        distance: Literal["jaccard", "cosine"] = "jaccard",
        feature_weights = None,
    ):
        self.out_dim = out_dim
        self.distance = distance
        if (self.distance == "jaccard"):
            self.compute_similarity = lambda x, y=None: internal.jaccard_similarity(x, y, feature_weights)
        elif (self.distance == "cosine"):
            self.compute_similarity = lambda x, y=None: internal.cosine_similarity(x, y, feature_weights)
        elif (self.distance == "rbf"):
            from sklearn.metrics.pairwise import rbf_kernel
            self.compute_similarity = lambda x, y=None: rbf_kernel(x, y)
        else:
            raise ValueError("Invalid distance")

    def fit(self, mat, verbose: int = 1):
        """
        mat
            Sparse matrix, note that if `distance == jaccard`, the matrix will be
            interpreted as a binary matrix.
        """
        self.sample = mat
        self.in_dim = mat.shape[1]
        if verbose > 0:
            logging.info("Compute similarity matrix")
        A = self.compute_similarity(mat)

        if (self.distance == "jaccard"):
            if verbose > 0:
                logging.info("Normalization")
            self.coverage = mat.sum(axis=1) / self.in_dim
            self.normalizer = JaccardNormalizer(A, self.coverage)
            self.normalizer.normalize(A, self.coverage, self.coverage)
            np.fill_diagonal(A, 0)
            # Remove outlier
            self.normalizer.outlier = np.quantile(A, 0.999)
            np.clip(A, a_min=0, a_max=self.normalizer.outlier, out=A)
        else:
            np.fill_diagonal(A, 0)

        # M <- D^-1/2 * A * D^-1/2
        D = np.sqrt(A.sum(axis=1)).reshape((-1, 1))
        np.divide(A, D, out=A)
        np.divide(A, D.T, out=A)

        if verbose > 0:
            logging.info("Perform decomposition")
        evals, evecs = sp.sparse.linalg.eigsh(A, self.out_dim, which='LM')
        ix = evals.argsort()[::-1]
        self.evals = np.real(evals[ix])
        self.evecs = np.real(evecs[:, ix])

        B = np.divide(self.evecs, D)
        np.divide(B, self.evals.reshape((1, -1)), out=B)

        self.B = B
        self.Q = []

        return self

    def extend(self, data):
        A = self.compute_similarity(self.sample, data)
        if (self.distance == "jaccard"):
            self.normalizer.normalize(
                A, self.coverage, data.sum(axis=1) / self.in_dim,
                clip_min=0, clip_max=self.normalizer.outlier
            )
        self.Q.append(A.T @ self.B)

    def transform(self, orthogonalize = True):
        if len(self.Q) > 0:
            Q = np.concatenate(self.Q, axis=0)
            D_ = np.sqrt(np.multiply(Q, self.evals.reshape(1, -1)) @ Q.sum(axis=0).T)
            np.divide(Q, D_.reshape((-1, 1)), out=Q)

            if orthogonalize:
                # orthogonalization
                sigma, V = np.linalg.eig(Q.T @ Q)
                sigma = np.sqrt(sigma)
                B = np.multiply(V.T, self.evals.reshape((1,-1))) @ V
                np.multiply(B, sigma.reshape((-1, 1)), out=B)
                np.multiply(B, sigma.reshape((1, -1)), out=B)
                evals_new, evecs_new = np.linalg.eig(B)

                # reorder
                ix = evals_new.argsort()[::-1]
                self.evals = evals_new[ix]
                evecs_new = evecs_new[:, ix]

                np.divide(evecs_new, sigma.reshape((-1, 1)), out=evecs_new)
                self.evecs = Q @ V @ evecs_new
            else:
                self.evecs = Q
        return (self.evals, self.evecs)

def orthogonalize(evals, evecs):
    _, sigma, Vt = np.linalg.svd(evecs, full_matrices=False)
    V = Vt.T

    B = np.multiply(V.T, evals.reshape((1,-1))) @ V
    np.multiply(B, sigma.reshape((-1, 1)), out=B)
    np.multiply(B, sigma.reshape((1, -1)), out=B)
    evals_new, evecs_new = np.linalg.eig(B)

    # reorder
    ix = evals_new.argsort()[::-1]
    evals_new = evals_new[ix]
    evecs_new = evecs_new[:, ix]

    np.divide(evecs_new, sigma.reshape((-1, 1)), out=evecs_new)
    evecs_new = evecs @ V @ evecs_new
    return (evals_new, evecs_new)

class JaccardNormalizer:
    def __init__(self, jm, c):
        (slope, intersect) = internal.jm_regress(jm, c)
        self.slope = slope
        self.intersect = intersect
        self.outlier = None

    def normalize(self, jm, c1, c2, clip_min=None, clip_max=None):
        # jm / (self.slope / (1 / c1 + 1 / c2.T - 1) + self.intersect)
        temp = 1 / c1 + 1 / c2.T
        temp -= 1
        np.reciprocal(temp, out=temp)
        np.multiply(temp, self.slope, out=temp)
        temp += self.intersect
        jm /= temp
        if clip_min is not None or clip_max is not None:
            np.clip(jm, a_min=clip_min, a_max=clip_max, out=jm)
        gc.collect()

def multi_spectral(
    adatas: list[internal.AnnData] | list[internal.AnnDataSet], 
    n_comps: int = 30,
    features: str | list[str] | list[np.ndarray] | None = "selected",
    weights: list[float] | None = None,
    random_state: int = 0,
    weighted_by_sd: bool = True,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Compute Laplacian Eigenmaps simultaneously on multiple modalities, with linear
    space and time complexity.

    This is similar to :func:`~snapatac2.tl.spectral`, but it can work on multiple modalities.

    Parameters
    ----------
    adatas
        A list of AnnData objects, representing single-cell data from different modalities.
    n_comps
        Number of dimensions to keep.
        See :func:`~snapatac2.tl.spectral` for details.
    features
        Boolean index mask. True means that the feature is kept. False means the feature is removed.
    weights
        Weights for each modality. If None, all modalities are weighted equally.
    random_state
        Seed of the random state generator
    weighted_by_sd
        Whether to weight the result eigenvectors by the square root of eigenvalues.
        See :func:`~snapatac2.tl.spectral` for details.

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        Return the eigenvalues and eigenvectors of the Laplacian matrix.

    See Also
    --------
    spectral
    """
    np.random.seed(random_state)

    if features is None or isinstance(features, str):
        features = [features] * len(adatas)
    if all(isinstance(f, str) for f in features):
        features = [adata.var[feature].to_numpy() for adata, feature in zip(adatas, features)]

    if weights is None:
        weights = [1.0 for _ in adatas]

    evals, evecs = internal.multi_spectral_embedding(adatas, features, weights, n_comps, random_state)

    if weighted_by_sd:
        idx = [i for i in range(evals.shape[0]) if evals[i] > 0]
        evals = evals[idx]
        evecs = evecs[:, idx] * np.sqrt(evals)

    return (evals, evecs)



from __future__ import annotations

import numpy as np
import itertools
from scipy.special import logsumexp

import snapatac2._snapatac2 as internal
from snapatac2._utils import is_anndata 

def mnc_correct(
    adata: internal.AnnData | internal.AnnDataSet | np.adarray,
    *,
    batch: str | list[str],
    n_neighbors: int = 5,
    n_clusters: int = 40,
    n_iter: int = 1,
    use_rep: str = "X_spectral",
    use_dims: int | list[int] | None = None,
    groupby: str | list[str] | None = None,
    key_added: str | None = None,
    inplace: bool = True,
    n_jobs: int = 8,
) -> np.ndarray | None:
    """
    A modified MNN-Correct algorithm based on cluster centroid.

    Parameters
    ----------
    data
        Matrice or AnnData object. Matrices should be shaped like n_obs x n_vars.
    batch
        Batch labels for cells, stored as a list of strings, where each string
        corresponds to the batch label for a cell. Alternatively, `batch` can be
        a string, and in this case labels will be obtained from `.obs[`batch`]`.
    n_neighbors
        Number of mutual nearest neighbors.
    n_clusters
        Number of clusters
    n_iter
        Number of iterations.
    use_rep
        Use the indicated representation in `.obsm`.
    use_dims
        Use these dimensions in `use_rep`.
    groupby
        If specified, split the data into groups and perform batch correction
        on each group separately.
    key_added
        If specified, add the result to ``adata.obsm`` with this key. Otherwise,
        it will be stored in ``adata.obsm[use_rep + "_mnn"]``.
    inplace
        Whether to store the result in the anndata object.
    n_jobs
        Number of jobs to use for parallelization.

    Returns
    -------
    np.ndarray | None
        if `inplace=True` it updates adata with the field
        ``adata.obsm[`use_rep`_mnn]``, containing adjusted principal components.
        Otherwise, it returns the result as a numpy array.
    """
    if is_anndata(adata):
        mat = adata.obsm[use_rep]
    else:
        mat = adata
        inplace = False

    if isinstance(use_dims, int): use_dims = range(use_dims) 
    mat = mat if use_dims is None else mat[:, use_dims]
    mat = np.asarray(mat)

    if isinstance(batch, str):
        batch = adata.obs[batch]
    elif isinstance(batch, list):
        assert len(batch) == mat.shape[0], "When `batch` is a list of strings,  \
            it is interpreted as the batch labels of cells and it must \
            have the same length as the number of cells."

    if groupby is None:
        mat = _mnc_correct_main(mat, batch, n_iter, n_neighbors, n_clusters)
    else:
        from multiprocess import Pool

        if isinstance(groupby, str): groupby = adata.obs[groupby]

        group_indices = {}
        for i, group in enumerate(groupby):
            if group in group_indices:
                group_indices[group].append(i)
            else:
                group_indices[group] = [i]
        group_indices = [x for x in group_indices.values()]

        inputs = [(mat[group_idx, :], batch[group_idx]) for group_idx in group_indices]
        with Pool(n_jobs) as p:
            results = p.map(lambda x: _mnc_correct_main(x[0], x[1], n_iter, n_neighbors, n_clusters), inputs)
        for idx, result in zip(group_indices, results):
            mat[idx, :] = result

    if inplace:
        if key_added is None:
            adata.obsm[use_rep + "_mnn"] = mat
        else:
            adata.obsm[key_added] = mat
    else:
        return mat

def _mnc_correct_main(
    data_matrix,
    batch_labels,
    n_iter,
    n_neighbors,
    n_clusters,
    random_state=0
):
    label_uniq = list(set(batch_labels))

    if len(label_uniq) > 1:
        for _ in range(n_iter):
            batch_idx = []
            data_by_batch = []
            for label in label_uniq:
                idx = [i for i, x in enumerate(batch_labels) if x == label]
                batch_idx.append(idx)
                data_by_batch.append(data_matrix[idx,:])
            new_matrix = _mnc_correct_multi(data_by_batch, n_neighbors, n_clusters, random_state)

            idx = list(itertools.chain.from_iterable(batch_idx))
            idx = np.argsort(idx)
            data_matrix = new_matrix[idx, :]
    return data_matrix

def _mnc_correct_multi(datas, n_neighbors, n_clusters, random_state):
    data0 = datas[0]
    for i in range(1, len(datas)):
        data1 = datas[i]
        pdata0, pdata1 = _mnc_correct_pair(data0, data1, n_clusters, n_neighbors, random_state)
        ratio = data0.shape[0] / (data0.shape[0] + data1.shape[0])
        data0_ = pdata0.project(data0, weight = 1 - ratio)
        data1_ = pdata1.project(data1, weight = ratio)
        data0 = np.concatenate((data0_, data1_), axis=0)
    return data0

def _mnc_correct_pair(X, Y, n_clusters, n_neighbors, random_state):
    from sklearn.neighbors import KDTree
    from sklearn.cluster import KMeans

    n_X = X.shape[0]
    n_Y = Y.shape[0]
    c_X = KMeans(
        n_clusters=min(n_clusters, n_X), n_init=10, random_state=random_state
    ).fit(X).cluster_centers_
    c_Y = KMeans(
        n_clusters=min(n_clusters, n_Y), n_init=10, random_state=random_state
    ).fit(Y).cluster_centers_

    tree_X = KDTree(c_X)
    tree_Y = KDTree(c_Y)

    # X by Y matrix
    m_X = tree_Y.query(c_X, k=min(n_neighbors, n_Y), return_distance=False)

    # Y by X matrix
    m_Y_ = tree_X.query(c_Y, k=min(n_neighbors, n_X), return_distance=False)
    m_Y = []
    for i in range(m_Y_.shape[0]):
        m_Y.append(set(m_Y_[i,:]))

    i_X = []
    i_Y = []
    for i in range(m_X.shape[0]):
        for j in m_X[i]:
            if i in m_Y[j]:
                i_X.append(i)
                i_Y.append(j)
    a = c_X[i_X,:]
    b = c_Y[i_Y,:]
    return (Projector(a, b), Projector(b, a))

class Projector(object):
    def __init__(self, X, Y):
        self.reference = X
        self.vector = Y - X

    def project(self, X, weight=0.5):
        def project(x):
            P = self.reference
            U = self.vector
            d = np.sqrt(np.sum((P - x)**2, axis=1))
            w = _normalize(-(d/0.005))
            #w = 1/d
            return (x + weight * np.average(U, axis=0, weights=w))
        return np.apply_along_axis(project, 1, X)

# exp transform the weights and then normalize them to sum to 1.
def _normalize(ws):
    s = logsumexp(ws)
    return np.exp(ws - s)


import snapatac2._snapatac2 as internal

import numpy as np
import pytest
from hypothesis import note, reproduce_failure, given, example, settings, HealthCheck, strategies as st
from hypothesis.extra.numpy import *
from sklearn.metrics import pairwise_distances
from scipy.sparse import csr_matrix
from scipy.spatial.distance import jaccard, cosine
from scipy.stats import pearsonr, spearmanr

def cosine_similarity(x, y, w = None):
    def dist(v1, v2):
        if np.amax(v1) == 0 or np.amax(v2) == 0:
            return 0
        else:
            return 1 - cosine(v1, v2, w)
    return pairwise_distances(x, y, metric = dist)

@given(
    mat1 =arrays(bool, (17, 50)),
    mat2 =arrays(bool, (19, 50)),
    w = arrays(
        np.float64, (50,),
        elements = {"allow_nan": False, "allow_infinity": False, "min_value": 1, "max_value": 1000},
    )
)
@settings(deadline = None, suppress_health_check = [HealthCheck.function_scoped_fixture])
def test_jaccard(mat1, mat2, w):
    csr1 = csr_matrix(mat1)
    csr2 = csr_matrix(mat2)

    np.testing.assert_array_equal(
        internal.jaccard_similarity(csr1),
        internal.jaccard_similarity(csr1, csr1),
    )

    np.testing.assert_array_almost_equal(
        pairwise_distances(mat1, mat1, metric= lambda x, y: 1 - jaccard(x, y)),
        internal.jaccard_similarity(csr1),
        decimal=12,
    )
    np.testing.assert_array_almost_equal(
        pairwise_distances(mat1, mat1, metric= lambda x, y: 1 - jaccard(x, y, w)),
        internal.jaccard_similarity(csr1, None, w),
        decimal=12,
    )

    np.testing.assert_array_almost_equal(
        pairwise_distances(mat1, mat2, metric= lambda x, y: 1 - jaccard(x, y)),
        internal.jaccard_similarity(csr1, csr2),
        decimal=12,
    )
    np.testing.assert_array_almost_equal(
        pairwise_distances(mat1, mat2, metric= lambda x, y: 1 - jaccard(x, y, w)),
        internal.jaccard_similarity(csr1, csr2, w),
        decimal=12,
    )

@given(
    mat1 = arrays(
        np.float64, (15, 50),
        elements = {"allow_nan": False, "allow_infinity": False, "min_value": 1, "max_value": 10000000},
    ),
    mat2 = arrays(
        np.float64, (17, 50),
        elements = {"allow_nan": False, "allow_infinity": False, "min_value": 1, "max_value": 10000000},
    ),
    w = arrays(
        np.float64, (50,),
        elements = {"allow_nan": False, "allow_infinity": False, "min_value": 1, "max_value": 1000},
    )
)
@settings(deadline = None, suppress_health_check = [HealthCheck.function_scoped_fixture])
def test_cosine(mat1, mat2, w):
    csr1 = csr_matrix(mat1)
    csr2 = csr_matrix(mat2)

    np.testing.assert_array_almost_equal(
        internal.cosine_similarity(csr1),
        internal.cosine_similarity(csr1, csr1),
        decimal=12,
    )

    np.testing.assert_array_almost_equal(
        cosine_similarity(mat1, mat1),
        internal.cosine_similarity(csr1),
        decimal=12,
    )
    np.testing.assert_array_almost_equal(
        cosine_similarity(mat1, mat1, w),
        internal.cosine_similarity(csr1, None, w),
        decimal=12,
    )

    np.testing.assert_array_almost_equal(
        cosine_similarity(mat1, mat2),
        internal.cosine_similarity(csr1, csr2),
        decimal=12,
    )
    np.testing.assert_array_almost_equal(
        cosine_similarity(mat1, mat2, w),
        internal.cosine_similarity(csr1, csr2, w),
        decimal=12,
    )

@given(
    mat1 = arrays(
        np.float64, (15, 50),
        elements = {"allow_subnormal": False, "allow_nan": False, "allow_infinity": False, "min_value": 1, "max_value": 10000000},
    ),
    mat2 = arrays(
        np.float64, (17, 50),
        elements = {"allow_subnormal": False, "allow_nan": False, "allow_infinity": False, "min_value": 1, "max_value": 10000000},
    ),
)
@settings(deadline = None, suppress_health_check = [HealthCheck.function_scoped_fixture])
def test_cor(mat1, mat2):
    def pearson(x, y):
        return pairwise_distances(x, y, metric=lambda a, b: pearsonr(a, b)[0])

    def spearman(x, y):
        return pairwise_distances(x, y, metric=lambda a, b: spearmanr(a, b)[0])

    np.testing.assert_array_almost_equal(
        spearman(mat1, mat2),
        internal.spearman(mat1, mat2),
        decimal=12,
    )


from __future__ import annotations

from typing import Literal
import numpy as np
from pathlib import Path
import logging

from snapatac2._snapatac2 import PyDNAMotif
from snapatac2._utils import fetch_seq
from snapatac2.genome import Genome
from snapatac2.tools._diff import _p_adjust_bh

def motif_enrichment(
    motifs: list[PyDNAMotif],
    regions: dict[str, list[str]],
    genome_fasta: Path | Genome,
    background: list[str] | None = None,
    method: Literal['binomial', 'hypergeometric'] | None = None,
) -> dict[str, 'polars.DataFrame']:
    """
    Identify enriched transcription factor motifs.

    Parameters
    ----------
    motifs
        A list of transcription factor motifs.
    regions
        Groups of regions. Each group will be tested independently against the background.
    genome_fasta
        A fasta file containing the genome sequences or a Genome object.
    background
        A list of regions to be used as the background. If None, the union of elements
        in `regions` will be used as the background.
    method
        Statistical testing method: "binomial" or "hypergeometric".
        To use "hypergeometric", the testing regions must be a subset of
        background regions.

    Returns
    -------
    dict[str, pl.DataFrame]:
        Dataframes containing the enrichment analysis results for different groups.
    """
    from pyfaidx import Fasta
    from tqdm import tqdm
    from scipy.stats import binom, hypergeom
    from math import log2
    import polars as pl

    def count_occurrence(query, idx_map, bound):
        return sum(bound[idx_map[q]] for q in query)

    if method is None:
        method = "hypergeometric" if background is None else "binomial"

    all_regions = set(p for ps in regions.values() for p in ps)
    if background is not None:
        for p in background:
            all_regions.add(p)
    all_regions = list(all_regions)
    region_to_idx = dict(map(lambda x: (x[1], x[0]), enumerate(all_regions)))

    logging.info("Fetching {} sequences ...".format(len(all_regions)))
    genome = genome_fasta.fasta if isinstance(genome_fasta, Genome) else str(genome_fasta)
    genome = Fasta(genome, one_based_attributes=False)
    sequences = [fetch_seq(genome, region) for region in all_regions]

    motif_id = []
    motif_name = []
    motif_family = []
    group_name = []
    fold_change = []
    n_fg = []
    N_fg = []
    n_bg = []
    N_bg = []
    logging.info("Computing enrichment ...")
    for motif in tqdm(motifs):
        bound = motif.with_nucl_prob().exists(sequences)
        if background is None:
            total_bg = len(bound)
            bound_bg = sum(bound)
        else:
            total_bg = len(background)
            bound_bg = count_occurrence(background, region_to_idx, bound)
        
        for key, val in regions.items():
            total_fg = len(val)
            bound_fg = count_occurrence(val, region_to_idx, bound)

            if bound_fg == 0:
                log_fc = 0 if bound_bg == 0 else float('-inf')
            else:
                log_fc = log2((bound_fg / total_fg) / (bound_bg / total_bg)) if bound_bg > 0 else float('inf')

            motif_id.append(motif.id)
            motif_name.append(motif.name)
            motif_family.append(motif.family)
            group_name.append(key)
            fold_change.append(log_fc)
            n_fg.append(bound_fg)
            N_fg.append(total_fg)
            n_bg.append(bound_bg)
            N_bg.append(total_bg)

          
    fold_change = np.array(fold_change)
    pval = np.zeros(len(fold_change))
    n_fg = np.array(n_fg)
    N_fg = np.array(N_fg)
    n_bg = np.array(n_bg)
    N_bg = np.array(N_bg)
    up_idx = fold_change >= 0
    down_idx = fold_change < 0
    if method == "binomial":
        pval[up_idx] = binom.sf(n_fg[up_idx] - 1, N_fg[up_idx], n_bg[up_idx] / N_bg[up_idx])
        pval[down_idx] = binom.cdf(n_fg[down_idx], N_fg[down_idx], n_bg[down_idx] / N_bg[down_idx])
    elif method == "hypergeometric":
        pval[up_idx] = hypergeom.sf(n_fg[up_idx] - 1, N_bg[up_idx], n_bg[up_idx], N_fg[up_idx])
        pval[down_idx] = hypergeom.cdf(n_fg[down_idx], N_bg[down_idx], n_bg[down_idx], N_fg[down_idx])
    else:
        raise NameError("'method' needs to be 'binomial' or 'hypergeometric'")
    pval = np.clip(pval, 1e-300, 1)

    result = dict(
        (key, {'id': [], 'name': [], 'family': [], 'log2(fold change)': [], 'p-value': []}) for key in regions.keys()
    )
    for i, key in enumerate(group_name):
        result[key]['id'].append(motif_id[i])
        result[key]['name'].append(motif_name[i])
        result[key]['family'].append(motif_family[i])
        result[key]['log2(fold change)'].append(fold_change[i])
        result[key]['p-value'].append(float(pval[i]))

    for key in result.keys():
        result[key]['adjusted p-value'] = _p_adjust_bh(result[key]['p-value'])
        result[key] = pl.DataFrame(result[key])
    return result


""" Implementation of the scrublet algorithm for single-cell ATAC-seq data
"""
from __future__ import annotations

import numpy as np
import scipy.sparse as ss
import logging
from anndata import AnnData

from .._utils import chunks, anndata_par
import snapatac2._snapatac2 as internal
from snapatac2.tools._embedding import spectral

def scrublet(
    adata: internal.AnnData | list[internal.AnnData],
    features: str | np.ndarray | None = "selected",
    n_comps: int = 15,
    sim_doublet_ratio: float = 2.0,
    expected_doublet_rate: float = 0.1,
    n_neighbors: int | None = None,
    use_approx_neighbors=False,
    random_state: int = 0,
    inplace: bool = True,
    n_jobs: int = 8,
    verbose: bool = True,
) -> None:
    """
    Compute probability of being a doublet using the scrublet algorithm.

    This function identifies doublets by generating simulated doublets using
    randomly pairing chromatin accessibility profiles of individual cells.
    The simulated doublets are then embedded alongside the original cells using
    the spectral embedding algorithm in this package.
    A k-nearest-neighbor classifier is trained to distinguish between the simulated
    doublets and the authentic cells.
    This trained classifier produces a "doublet score" for each cell.
    The doublet scores are then converted into probabilities using a Gaussian mixture model.

    Parameters
    ----------
    adata
        The (annotated) data matrix of shape `n_obs` x `n_vars`.
        Rows correspond to cells and columns to regions.
        `adata` can also be a list of AnnData objects.
        In this case, the function will be applied to each AnnData object in parallel.
    features
        Boolean index mask, where `True` means that the feature is kept, and
        `False` means the feature is removed.
    n_comps
        Number of components. 15 is usually sufficient. The algorithm is not sensitive
        to this parameter.
    sim_doublet_ratio
        Number of doublets to simulate relative to the number of observed cells.
    expected_doublet_rate
        Expected doublet rate.
    n_neighbors
        Number of neighbors used to construct the KNN graph of observed
        cells and simulated doublets. If `None`, this is 
        set to round(0.5 * sqrt(n_cells))
    use_approx_neighbors
        Whether to use approximate search.
    random_state
        Random state.
    inplace
        Whether update the AnnData object inplace
    n_jobs
        Number of jobs to run in parallel.
    verbose
        Whether to print progress messages.
    
    Returns
    -------
    tuple[np.ndarray, np.ndarray] | None:
        if ``inplace = True``, it updates adata with the following fields:
            - ``adata.obs["doublet_probability"]``: probability of being a doublet
            - ``adata.obs["doublet_score"]``: doublet score
    """
    if isinstance(adata, list):
        result = anndata_par(
            adata,
            lambda x: scrublet(x, features, n_comps, sim_doublet_ratio,
                               expected_doublet_rate, n_neighbors,
                               use_approx_neighbors, random_state,
                               inplace, n_jobs, verbose=False),
            n_jobs=n_jobs,
        )
        if inplace:
            return None
        else:
            return result

    if isinstance(features, str):
        if features in adata.var:
            features = adata.var[features].to_numpy()
        else:
            raise NameError("Please call `select_features` first or explicitly set `features = None`")

    if features is None:
        count_matrix = adata.X[:]
    else:
        count_matrix = adata.X[:, features]

    if min(count_matrix.shape) == 0: raise NameError("Matrix is empty")

    if n_neighbors is None:
        n_neighbors = int(round(0.5 * np.sqrt(count_matrix.shape[0])))

    doublet_scores_obs, doublet_scores_sim, _, _ = scrub_doublets_core(
        count_matrix, n_neighbors, sim_doublet_ratio, expected_doublet_rate,
        n_comps=n_comps,
        use_approx_neighbors = use_approx_neighbors,
        random_state=random_state,
        verbose=verbose,
    )
    probs = get_doublet_probability(
        doublet_scores_sim, doublet_scores_obs, random_state,
    )
 
    if inplace:
        adata.obs["doublet_probability"] = probs
        adata.obs["doublet_score"] = doublet_scores_obs
        adata.uns["scrublet_sim_doublet_score"] = doublet_scores_sim
    else:
        return probs, doublet_scores_obs

def filter_doublets(
    adata: internal.AnnData | list[internal.AnnData],
    probability_threshold: float | None = 0.5,
    score_threshold: float | None = None,
    inplace: bool = True,
    n_jobs: int = 8,
    verbose: bool = True,
) -> np.ndarray | None:
    """Remove doublets according to the doublet probability or doublet score.

    The user can choose to remove doublets by either the doublet probability or the doublet score.
    :func:`~snapatac2.pp.scrublet` must be ran first in order to use this function.

    Parameters
    ----------
    adata
        The (annotated) data matrix of shape `n_obs` x `n_vars`.
        Rows correspond to cells and columns to regions.
    probability_threshold
        Threshold for doublet probability. Doublet probability greater than
        this threshold will be removed. The default value is 0.5. Using a lower
        threshold will remove more cells.
    score_threshold
        Threshold for doublet score. Doublet score greater than this threshold
        will be removed. Only one of `probability_threshold` and `score_threshold`
        can be set. Using `score_threshold` is not recommended for most cases.
    inplace
        Perform computation inplace or return result.
    n_jobs
        Number of jobs to run in parallel.
    verbose
        Whether to print progress messages.

    Returns
    -------
    np.ndarray | None:
        If `inplace = True`, directly subsets the data matrix. Otherwise return 
        a boolean index mask that does filtering, where `True` means that the
        cell is kept, `False` means the cell is removed.

    See Also
    --------
    scrublet
    """
    if isinstance(adata, list):
        result = anndata_par(
            adata,
            lambda x: filter_doublets(x, probability_threshold, score_threshold,
                                      inplace, n_jobs, verbose=False),
            n_jobs=n_jobs,
        )
        if inplace:
            return None
        else:
            return result

    if probability_threshold is not None and score_threshold is not None:
        raise ValueError("Only one of `probability_threshold` and `score_threshold` can be set.")
    if probability_threshold is not None:
        probs = adata.obs["doublet_probability"].to_numpy()
        is_doublet = probs > probability_threshold
    if score_threshold is not None:
        scores = adata.obs["doublet_score"].to_numpy()
        is_doublet = scores > score_threshold

    doublet_rate = np.mean(is_doublet)
    if verbose: logging.info(f"Detected doublet rate = {doublet_rate*100:.3f}%")

    if inplace:
        adata.uns["doublet_rate"] = doublet_rate
        if adata.isbacked:
            adata.subset(~is_doublet)
        else:
            adata._inplace_subset_obs(~is_doublet)
    else:
        return ~is_doublet

def scrub_doublets_core(
    count_matrix: ss.spmatrix,
    n_neighbors: int,
    sim_doublet_ratio: float,
    expected_doublet_rate: float,
    synthetic_doublet_umi_subsampling: float =1.0,
    n_comps: int = 30,
    use_approx_neighbors: bool = False,
    random_state: int = 0,
    verbose: bool = False,
) -> None:
    """
    Modified scrublet pipeline for single-cell ATAC-seq data.

    Automatically sets a threshold for calling doublets, but it's best to check 
    this by running plot_histogram() afterwards and adjusting threshold 
    with call_doublets(threshold=new_threshold) if necessary.

    Arguments
    ---------
    synthetic_doublet_umi_subsampling
        Rate for sampling UMIs when creating synthetic doublets. If 1.0, 
        each doublet is created by simply adding the UMIs from two randomly 
        sampled observed transcriptomes. For values less than 1, the 
        UMI counts are added and then randomly sampled at the specified
        rate.
    n_comps
        Number of principal components used to embed the transcriptomes prior
        to k-nearest-neighbor graph construction.

    Returns
    -------
    doublet_scores_obs_, doublet_errors_obs_,
    doublet_scores_sim_, doublet_errors_sim_,
    predicted_doublets_, z_scores_ 
    threshold_, detected_doublet_rate_,
    detectable_doublet_fraction_, overall_doublet_rate_,
    doublet_parents_, doublet_neighbor_parents_ 
    """
    import gc

    total_counts_obs = count_matrix.sum(1).A.squeeze()

    if verbose: logging.info('Simulating doublets...')
    (count_matrix_sim, total_counts_sim, _) = simulate_doublets(
        count_matrix, total_counts_obs, sim_doublet_ratio,
        synthetic_doublet_umi_subsampling, random_state
    )

    if verbose: logging.info('Spectral embedding ...')
    n = count_matrix.shape[0]
    merged_matrix = ss.vstack([count_matrix, count_matrix_sim])
    del count_matrix_sim
    gc.collect()
    _, evecs = spectral(
        AnnData(X=merged_matrix),
        features=None,
        n_comps=n_comps,
        inplace=False,
    )
    manifold = np.asanyarray(evecs)
    manifold_obs = manifold[0:n, ]
    manifold_sim = manifold[n:, ]

    if verbose: logging.info('Calculating doublet scores...')
    doublet_scores_obs, doublet_scores_sim = calculate_doublet_scores(
        manifold_obs, manifold_sim, k = n_neighbors,
        exp_doub_rate = expected_doublet_rate,
        use_approx_neighbors = use_approx_neighbors,
        random_state = random_state,
    )

    return (doublet_scores_obs, doublet_scores_sim, manifold_obs, manifold_sim)


def simulate_doublets(
    count_matrix: ss.spmatrix,
    total_counts: np.ndarray,
    sim_doublet_ratio: int = 2,
    synthetic_doublet_umi_subsampling: float = 1.0,
    random_state: int = 0,
) -> tuple[ss.spmatrix, np.ndarray, np.ndarray]:
    """
    Simulate doublets by adding the counts of random cell pairs.

    Parameters
    ----------
    count_matrix
    total_counts
        Total insertion counts in each cell
    sim_doublet_ratio
        Number of doublets to simulate relative to the number of cells.
    synthetic_doublet_umi_subsampling
        Rate for sampling UMIs when creating synthetic doublets. If 1.0, 
        each doublet is created by simply adding the UMIs from two randomly 
        sampled observed transcriptomes. For values less than 1, the 
        UMI counts are added and then randomly sampled at the specified rate.

    Returns
    -------
    count_matrix_sim

    """
    n_obs = count_matrix.shape[0]
    n_sim = int(n_obs * sim_doublet_ratio)

    np.random.seed(random_state)
    pair_ix = np.random.randint(0, n_obs, size=(n_sim, 2))

    count_matrix_sim = count_matrix[pair_ix[:,0],:] + count_matrix[pair_ix[:,1],:]
    total_counts_sim = total_counts[pair_ix[:,0]] + total_counts[pair_ix[:,1]]

    if synthetic_doublet_umi_subsampling < 1:
        pass
        #count_matrix_sim, total_counts_sim = subsample_counts(
        #    count_matrix_sim, synthetic_doublet_umi_subsampling, total_counts_sim,
        #    random_seed=random_state
        #)
    return (count_matrix_sim, total_counts_sim, pair_ix)

def calculate_doublet_scores(
    manifold_obs: np.ndarray,
    manifold_sim: np.ndarray,
    k: int = 40,
    exp_doub_rate: float = 0.1,
    stdev_doub_rate: float = 0.03,
    use_approx_neighbors=False,
    random_state: int = 0,
) -> None:
    """
    Parameters
    ----------
    manifold_obs
        Manifold of observations
    manifold_sim
        Manifold of simulated doublets
    k
        Number of nearest neighbors
    exp_doub_rate
    stdev_doub_rate
    random_state
    """
    from sklearn.neighbors import NearestNeighbors

    n_obs = manifold_obs.shape[0]
    n_sim = manifold_sim.shape[0]

    manifold = np.vstack((manifold_obs, manifold_sim))
    doub_labels = np.concatenate(
        (np.zeros(n_obs, dtype=int), np.ones(n_sim, dtype=int))
    )

    # Adjust k (number of nearest neighbors) based on the ratio of simulated to observed cells
    k_adj = int(round(k * (1 + n_sim / float(n_obs))))
    
    # Find k_adj nearest neighbors
    if use_approx_neighbors:
        knn = internal.approximate_nearest_neighbour_graph(
            manifold.astype(np.float32), k_adj)
    else:
        knn = internal.nearest_neighbour_graph(manifold, k_adj)
    indices = knn.indices
    indptr = knn.indptr
    neighbors = np.vstack(
        [indices[indptr[i]:indptr[i+1]] for i in range(len(indptr) - 1)]
    )
    
    # Calculate doublet score based on ratio of simulated cell neighbors vs. observed cell neighbors
    doub_neigh_mask = doub_labels[neighbors] == 1
    n_sim_neigh = doub_neigh_mask.sum(1)
    n_obs_neigh = doub_neigh_mask.shape[1] - n_sim_neigh
    
    rho = exp_doub_rate
    r = n_sim / float(n_obs)
    nd = n_sim_neigh.astype(float)
    ns = n_obs_neigh.astype(float)
    N = float(k_adj)
    
    # Bayesian
    q=(nd+1)/(N+2)
    Ld = q*rho/r/(1-rho-q*(1-rho-rho/r))

    se_q = np.sqrt(q*(1-q)/(N+3))
    se_rho = stdev_doub_rate

    se_Ld = q*rho/r / (1-rho-q*(1-rho-rho/r))**2 * np.sqrt((se_q/q*(1-rho))**2 + (se_rho/rho*(1-q))**2)

    doublet_scores_obs = Ld[doub_labels == 0]
    doublet_scores_sim = Ld[doub_labels == 1]
    doublet_errors_obs = se_Ld[doub_labels==0]
    doublet_errors_sim = se_Ld[doub_labels==1]

    return (doublet_scores_obs, doublet_scores_sim)

def get_doublet_probability(
    doublet_scores_sim: np.ndarray,
    doublet_scores: np.ndarray,
    random_state: int = 0,
    verbose: bool = False,
):
    from sklearn.mixture import BayesianGaussianMixture

    X = doublet_scores_sim.reshape((-1, 1))
    gmm = BayesianGaussianMixture(
        n_components=2, n_init=10, max_iter=1000, random_state=random_state
    ).fit(X)

    if verbose:
        logging.info("GMM means: {}".format(gmm.means_))

    i = np.argmax(gmm.means_)
    return gmm.predict_proba(doublet_scores.reshape((-1, 1)))[:,i]



from __future__ import annotations

from typing import Literal
from pathlib import Path

import snapatac2._snapatac2 as internal
from snapatac2._utils import get_file_format

def export_fragments(
    adata: internal.AnnData | internal.AnnDataSet,
    groupby: str | list[str],
    selections: list[str] | None = None,
    ids: str | list[str] | None = None,
    min_frag_length: int | None = None,
    max_frag_length: int | None = None,
    out_dir: Path = "./",
    prefix: str = "",
    suffix: str = ".bed.zst",
    compression: Literal["gzip", "zstandard"] | None = None,
    compression_level: int | None = None,
) -> dict[str, str]:
    """Export and save fragments in a BED format file.

    Parameters
    ----------
    adata
        The (annotated) data matrix of shape `n_obs` x `n_vars`.
        Rows correspond to cells and columns to regions.
    groupby
        Group the cells. If a list of `str`, each element is the group name of the corresponding cell.
        The length of the list must be equal to `n_obs`. If a `str`, groups are obtained from
        `.obs[groupby]`. 
    selections
        Export only the selected groups.
    ids
        Cell ids add to the bed records. If `None`, `.obs_names` is used.
    min_frag_length
        Minimum fragment length to be included in the computation.
    max_frag_length
        Maximum fragment length to be included in the computation.
    out_dir
        Directory for saving the outputs.
    prefix
        Text added to the output file name.
    suffix
        Text added to the output file name.
    compression
        Compression type. If `None`, it is inferred from the suffix.
    compression_level
        Compression level. 1-9 for gzip, 1-22 for zstandard.
        If `None`, it is set to 6 for gzip and 3 for zstandard.

    Returns
    -------
    dict[str, str]
        A dictionary contains `(groupname, filename)` pairs. The file names are
        formatted as `{prefix}{groupname}{suffix}`.

    See Also
    --------
    export_coverage
    """
    if isinstance(groupby, str):
        groupby = adata.obs[groupby]
    if selections is not None:
        selections = set(selections)
    
    if ids is None:
        ids = adata.obs_names
    elif isinstance(ids, str):
        ids = adata.obs[ids]

    if compression is None:
        _, compression = get_file_format(suffix)

    return internal.export_fragments(
        adata, list(ids), list(groupby), out_dir, prefix, suffix, selections, 
        min_frag_length, max_frag_length, compression, compression_level,
    )

def export_coverage(
    adata: internal.AnnData | internal.AnnDataSet,
    groupby: str | list[str],
    selections: list[str] | None = None,
    bin_size: int = 10,
    blacklist: Path | None = None,
    normalization: Literal["RPKM", "CPM", "BPM"] | None = "RPKM",
    include_for_norm: list[str] | Path = None,
    exclude_for_norm: list[str] | Path = None,
    min_frag_length: int | None = None,
    max_frag_length: int | None = 2000,
    counting_strategy: Literal['fragment', 'insertion'] = 'fragment',
    smooth_base: int | None = None,
    out_dir: Path = "./",
    prefix: str = "",
    suffix: str = ".bw",
    output_format: Literal["bedgraph", "bigwig"] | None = None,
    compression: Literal["gzip", "zstandard"] | None = None,
    compression_level: int | None = None,
    tempdir: Path | None = None,
    n_jobs: int = 8,
) -> dict[str, str]:
    """Export and save coverage in a bedgraph or bigwig format file.

    This function first divides cells into groups based on the `groupby` parameter.
    It then independently generates the genome-wide coverage track (bigWig or bedGraph) for each group
    of cells. The coverage is calculated as the number of reads
    per bin, where bins are short consecutive counting windows of a defined size.
    For paired-end data, the reads are extended to the fragment length and the
    coverage is calculated as the number of fragments per bin.
    There are several options for normalization. The default is RPKM, which
    normalizes by the total number of reads and the length of the region.
    The normalization can be disabled by setting `normalization=None`.

    .. image:: /_static/images/func+export_coverage.svg
        :align: center

    Parameters
    ----------
    adata
        The (annotated) data matrix of shape `n_obs` x `n_vars`.
        Rows correspond to cells and columns to regions.
    groupby
        Group the cells. If a list of `str`, each element is the group name of the corresponding cell.
        The length of the list must be equal to `n_obs`. If a `str`, groups are obtained from
        `.obs[groupby]`. 
    selections
        Export only the selected groups.
    bin_size
        Size of the bins, in bases, for the output of the bigwig/bedgraph file.
    blacklist
        A BED file containing the blacklisted regions.
    normalization
        Normalization method. If `None`, no normalization is performed. Options:
        - RPKM (per bin) = #reads per bin / (#mapped_reads (in millions) * bin length (kb)).
        - CPM (per bin) = #reads per bin / #mapped_reads (in millions).
        - BPM (per bin) = #reads per bin / sum of all reads per bin (in millions).
    include_for_norm
        A list of string (e.g., ["chr1:1-100", "chr2:2-200"]) or a BED file containing
        the genomic loci to include for normalization.
        If specified, only the reads that overlap with these loci will be used for normalization.
        A typical use case is to include only the promoter regions for the normalization.
    exclude_for_norm
        A list of string (e.g., ["chr1:1-100", "chr2:2-200"]) or a BED file containing
        the genomic loci to exclude for normalization.
        If specified, the reads that overlap with these loci will be excluded from normalization.
        If a read overlaps with both `include_for_norm` and `exclude_for_norm`, it will be excluded.
    min_frag_length
        Minimum fragment length to be included in the computation.
    max_frag_length
        Maximum fragment length to be included in the computation.
    counting_strategy
        The strategy to compute feature counts. It must be one of the following:
        "fragment" or "insertion". "fragment" means the
        feature counts are assigned based on the number of fragments that overlap
        with a region of interest. "insertion" means the feature counts are assigned
        based on the number of insertions that overlap with a region of interest.
    smooth_base
        Length of the smoothing window in bases for the output of the bigwig/bedgraph file.
    out_dir
        Directory for saving the outputs.
    prefix
        Text added to the output file name.
    suffix
        Text added to the output file name.
    output_format
        Output format. If `None`, it is inferred from the suffix.
    compression
        Compression type. If `None`, it is inferred from the suffix.
    compression_level
        Compression level. 1-9 for gzip, 1-22 for zstandard.
        If `None`, it is set to 6 for gzip and 3 for zstandard.
    n_jobs
        Number of threads to use. If `<= 0`, use all available threads.

    Returns
    -------
    dict[str, str]
        A dictionary contains `(groupname, filename)` pairs. The file names are
        formatted as `{prefix}{groupname}{suffix}`.

    See Also
    --------
    export_fragments

    Examples
    --------
    >>> import snapatac2 as snap
    >>> data = snap.read(snap.datasets.pbmc5k(type="annotated_h5ad"), backed='r')
    >>> snap.ex.export_coverage(data, groupby='cell_type', suffix='.bedgraph.zst')
    {'cDC': './cDC.bedgraph.zst',
     'Memory B': './Memory B.bedgraph.zst',
     'CD4 Naive': './CD4 Naive.bedgraph.zst',
     'pDC': './pDC.bedgraph.zst',
     'CD8 Naive': './CD8 Naive.bedgraph.zst',
     'CD8 Memory': './CD8 Memory.bedgraph.zst',
     'CD14 Mono': './CD14 Mono.bedgraph.zst',
     'Naive B': './Naive B.bedgraph.zst',
     'NK': './NK.bedgraph.zst',
     'CD4 Memory': './CD4 Memory.bedgraph.zst',
     'CD16 Mono': './CD16 Mono.bedgraph.zst',
     'MAIT': './MAIT.bedgraph.zst'}
    >>> snap.ex.export_coverage(data, groupby='cell_type', suffix='.bw')
    {'Naive B': './Naive B.bw',
     'CD4 Memory': './CD4 Memory.bw',
     'CD16 Mono': './CD16 Mono.bw',
     'CD8 Naive': './CD8 Naive.bw',
     'pDC': './pDC.bw',
     'CD8 Memory': './CD8 Memory.bw',
     'NK': './NK.bw',
     'Memory B': './Memory B.bw',
     'CD14 Mono': './CD14 Mono.bw',
     'MAIT': './MAIT.bw',
     'CD4 Naive': './CD4 Naive.bw',
     'cDC': './cDC.bw'}
    """
    if isinstance(groupby, str):
        groupby = adata.obs[groupby]
    if selections is not None:
        selections = set(selections)
    
    if output_format is None:
        output_format, inferred_compression = get_file_format(suffix)
        if output_format is None:
            raise ValueError("Output format cannot be inferred from suffix.")
        if compression is None:
            compression = inferred_compression

    n_jobs = None if n_jobs <= 0 else n_jobs
    return internal.export_coverage(
        adata, list(groupby), bin_size, out_dir, prefix, suffix, output_format, counting_strategy,
        selections, blacklist, normalization, include_for_norm, exclude_for_norm, min_frag_length,
        max_frag_length, smooth_base, compression, compression_level, tempdir, n_jobs,
    )


from __future__ import annotations

from pathlib import Path
from anndata import AnnData
import snapatac2._snapatac2 as internal
from scipy.sparse import csr_matrix

def read_10x_mtx(
    path: Path,
    file: Path | None = None,
    prefix: str | None = None,
) -> AnnData:
    """Read 10x-Genomics-formatted mtx directory.

    Parameters
    ----------
    path
        Path to directory for `.mtx` and `.tsv` files. The directory should contain
        three files:

        1. count matrix: "matrix.mtx" or "matrix.mtx.gz".
        2. features: "genes.tsv", or "genes.tsv.gz", or "features.tsv", or "features.tsv.gz".
        3. barcodes: "barcodes.tsv", or "barcodes.tsv.gz".
    file
        File name of the ".h5ad" file used to save the AnnData object. If `None`,
        an in-memory AnnData object is returned.
    prefix
        Any prefix before `matrix.mtx`, `genes.tsv` and `barcodes.tsv`. For instance,
        if the files are named `patientA_matrix.mtx`, `patientA_genes.tsv` and
        `patientA_barcodes.tsv`, then the prefix is `patientA_`.

    Returns
    -------
    AnnData
        An AnnData object.
    """
    import pandas as pd

    def get_files(prefix, names):
        return list(filter(
            lambda x: x.is_file(),
            map(lambda x: Path(path + "/" + prefix + x), names)
        ))

    prefix = "" if prefix is None else prefix

    matrix_files = get_files(prefix, ["matrix.mtx", "matrix.mtx.gz"])
    n = len(matrix_files)
    if n == 1:
        mat = csr_matrix(internal.read_mtx(str(matrix_files[0])).X[:].T)
        adata = AnnData(X=mat) if file is None else internal.AnnData(X=mat, filename=file)
    else:
        raise ValueError("Expecting a single 'matrix.mtx' or 'matrix.mtx.gz' file, but found {}.".format(n))

    feature_files = get_files(
        prefix,
        ["genes.tsv", "genes.tsv.gz", "features.tsv", "features.tsv.gz"]
    )
    n = len(feature_files)
    if n == 1:
        df = pd.read_csv(str(feature_files[0]), sep='\t', header=None, index_col=0)
        df.index.name = "index"
        adata.var_names = df.index
        adata.var = df
    else:
        raise ValueError("Expecting a single feature file, but found {}.".format(n))

    barcode_files = get_files(prefix, ["barcodes.tsv", "barcodes.tsv.gz"])
    n = len(barcode_files)
    if n == 1:
        df = pd.read_csv(str(barcode_files[0]), sep='\t', header=None, index_col=0)
        df.index.name = "index"
        adata.obs_names = df.index
        adata.obs = df
    else:
        raise ValueError("Expecting a single barcode file, but found {}.".format(n))

    return adata